local nn = require 'nn'
local cunn = require 'cunn'
local cudnn = require 'cudnn'
local nngraph = require 'nngraph'
local nnlib = cudnn

local Residual = require('src.models.layers.Residual')

local function hourglass(n, f, inp, nResidual)
  -- Upper branch
  local up1 = inp
  for i = 1,nResidual do up1 = Residual(f,f)(up1) end

  -- Lower branch
  local low1 = nnlib.SpatialMaxPooling(2,2,2,2)(inp)
  for i = 1,nResidual do low1 = Residual(f,f)(low1) end
  local low2

  if n > 1 then low2 = hourglass(n-1,f,low1, nResidual)
  else
    low2 = low1
    for i = 1,nResidual do low2 = Residual(f,f)(low2) end
  end

  local low3 = low2
  for i = 1,nResidual do low3 = Residual(f,f)(low3) end
  local up2 = nn.SpatialUpSamplingNearest(2)(low3)

  -- Bring two branches together
  return nn.CAddTable()({up1,up2})
end

local function lin(numIn,numOut,inp)
  -- Apply 1x1 convolution, stride 1, no padding
  local l = nnlib.SpatialConvolution(numIn,numOut,1,1,1,1,0,0)(inp)
  return nnlib.ReLU(true)(nn.SpatialBatchNormalization(numOut)(l))
end

function createModel(opt)
  -- load compositional structure of human bodies
  local struct = opt.struct
  local nSemanticLevels = opt.nSemanticLevels
  local nParts = struct.nParts 
  assert(nSemanticLevels <= #nParts)

  local nStack = opt.nStack

  local nResidual = opt.nResidual -- =1
  local inp = nn.Identity()()

  -- Initial processing of the image
  local cnv1_ = nnlib.SpatialConvolution(3,64,3,3,1,1,1,1)(inp)           -- 256
  local cnv1 = nnlib.ReLU(true)(nn.SpatialBatchNormalization(64)(cnv1_))
  local pool1 = nnlib.SpatialMaxPooling(2,2,2,2)(cnv1)                    -- 128 
  local cnv2_ = nnlib.SpatialConvolution(64,64,3,3,1,1,1,1)(pool1)           
  local cnv2 = nnlib.ReLU(true)(nn.SpatialBatchNormalization(64)(cnv2_))
  local cnv3_ = nnlib.SpatialConvolution(64,64,3,3,1,1,1,1)(cnv2)
  local cnv3 = nnlib.ReLU(true)(nn.SpatialBatchNormalization(64)(cnv3_))

  local r1 = Residual(64,128)(cnv3)
  local pool = nnlib.SpatialMaxPooling(2,2,2,2)(r1)                       -- 64
  local r4 = Residual(128,128)(pool)
  local r5 = Residual(128,opt.nFeats)(r4)

  local out = {}
  local inter = r5

  -- Bottom-up inference 
  local outIdxSet = {}
  for levelIdx = 1,nSemanticLevels do
    local hg = hourglass(4,opt.nFeats,inter, nResidual)

    -- Residual layers at output resolution
    local ll = hg
    -- for j = 1,nResidual do ll = Residual(opt.nFeats,opt.nFeats)(ll) end
    -- Linear layer to produce first set of predictions
    ll = lin(opt.nFeats,opt.nFeats,ll)

    -- Predicted heatmaps
    local nOutChannels = nParts[levelIdx]
    if opt.catPartEnds and levelIdx ~= 1 then
      nOutChannels = nOutChannels + nParts[1]
    end

    local tmpOut = nnlib.SpatialConvolution(opt.nFeats,nOutChannels,
        1,1,1,1,0,0)(ll)
    table.insert(out,tmpOut)

    -- Add predictions back if this is not the last hg module
    if #out < nStack then
      local ll_ = nnlib.SpatialConvolution(opt.nFeats,opt.nFeats,
          1,1,1,1,0,0)(ll)
      local tmpOut_ = nnlib.SpatialConvolution(nOutChannels,opt.nFeats,
          1,1,1,1,0,0)(tmpOut)
      inter = nn.CAddTable()({inter, ll_, tmpOut_})
    end

    table.insert(outIdxSet, 1, #out)
  end

  table.remove(outIdxSet, 1)
  -- Top-down inference
  for levelIdx = nSemanticLevels-1,1,-1 do
    -- Top-down connections
    local nRemainChannels = opt.nFeats - nParts[levelIdx]
    if opt.catPartEnds and levelIdx ~= 1 then
      nRemainChannels = nRemainChannels - nParts[1]
    end
    local inter_ = Residual(opt.nFeats, nRemainChannels)(inter)
    local srcOutIdx = table.remove(outIdxSet, 1)
    inter = nn.JoinTable(2)({out[srcOutIdx], inter_})

    local hg = hourglass(4,opt.nFeats,inter, nResidual)

    -- Residual layers at output resolution
    local ll = hg
    -- for j = 1,nResidual do ll = Residual(opt.nFeats,opt.nFeats)(ll) end
    -- Linear layer to produce first set of predictions
    ll = lin(opt.nFeats,opt.nFeats,ll)

    -- Predicted heatmaps
    nOutChannels = nParts[levelIdx]
    if opt.catPartEnds and levelIdx ~= 1 then
      nOutChannels = nOutChannels + nParts[1]
    end
    local tmpOut = nnlib.SpatialConvolution(opt.nFeats,nOutChannels,
        1,1,1,1,0,0)(ll)
    table.insert(out,tmpOut)

    -- Add predictions back if this is not the last hg module
    if #out < nStack then
      local ll_ = nnlib.SpatialConvolution(opt.nFeats,opt.nFeats,
          1,1,1,1,0,0)(ll)
      local tmpOut_ = nnlib.SpatialConvolution(nOutChannels,opt.nFeats,
          1,1,1,1,0,0)(tmpOut)
      inter = nn.CAddTable()({inter, ll_, tmpOut_})
    end
  end

  -- Final model
  local model = nn.gModule({inp}, out)

  -- return model
  return model:cuda()

end

return createModel
